#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <xmms/plugin.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "preferences.h"
#include "config.h"

static EffectPlugin compressor;
static CompressorPrefs prefs;
static void myInit(void);
static void myCleanup(void);
static int myModify(gpointer *, gint, AFormat, gint, gint);
static void myAbout(void);
static void myPrefs(void);

/* Monitor and preferences dialog */
static Display *display;
static Window window;
static Visual *visual;
static int screen;
static GC blackGC, whiteGC, blueGC, yellowGC, dkyellowGC, redGC;
static int *peaks = NULL;

EffectPlugin *get_eplugin_info(void)
{
	memset(&compressor, 0, sizeof(compressor));

	compressor.description = "KOMPRESSOR CRUSH XMMS";
	compressor.init = myInit;
	compressor.cleanup = myCleanup;
	compressor.mod_samples = myModify;
	compressor.about = myAbout;
	compressor.configure = myPrefs;

	return &compressor;
}

static int gainCurrent, gainTarget;

void AllocPeaks(void)
{
	static int lastsize = 0;

	peaks = realloc(peaks, sizeof(int)*prefs.buckets);

	if (prefs.buckets > lastsize)
		memset(peaks + lastsize, 0, sizeof(int)*(prefs.buckets
							 - lastsize));
	XResizeWindow(display, window, prefs.buckets, 128 + 8);
	XFlush(display);

	lastsize = prefs.buckets;
}

void setMonitorVisibility(gboolean visible)
{
	
	if (visible) {
		XMapWindow(display, window);
	} else
	{
		XUnmapWindow(display, window);
	}
	
	XFlush(display);
}

void myInit(void)
{
	
	XGCValues gcv;
	XColor col;
	
	initPrefs(&prefs);
	
	gainCurrent = gainTarget = (1 << GAINSHIFT);

	display = XOpenDisplay(getenv("DISPLAY"));
	screen = DefaultScreen(display);
	visual = DefaultVisual(display, screen);
	window = XCreateSimpleWindow(display,
				     RootWindow(display, screen),
				     0, 0, prefs.buckets, 128 + 8, 0,
				     BlackPixel(display, screen),
				     WhitePixel(display, screen));
	XStoreName(display, window, "xmms-compress monitor");
	
	gcv.foreground = BlackPixel(display, screen);
	blackGC = XCreateGC(display, window, GCForeground, &gcv);
	gcv.foreground = WhitePixel(display, screen);
	whiteGC = XCreateGC(display, window, GCForeground, &gcv);
	col.red = 0;
	col.green = 0;
	col.blue = 65535;
	XAllocColor(display, DefaultColormap(display, screen), &col);
	gcv.foreground = col.pixel;
	blueGC = XCreateGC(display, window, GCForeground, &gcv);
	col.red = 65535;
	col.green = 65535;
	col.blue = 0;
	XAllocColor(display, DefaultColormap(display, screen), &col);
	gcv.foreground = col.pixel;
	yellowGC = XCreateGC(display, window, GCForeground, &gcv);
	col.red = 32767;
	col.green = 32767;
	col.blue = 0;
	XAllocColor(display, DefaultColormap(display, screen), &col);
	gcv.foreground = col.pixel;
	dkyellowGC = XCreateGC(display, window, GCForeground, &gcv);
	col.red = 65535;
	col.green = 0;
	col.blue = 0;
	XAllocColor(display, DefaultColormap(display, screen), &col);
	gcv.foreground = col.pixel;
	redGC = XCreateGC(display, window, GCForeground, &gcv);

	/* Allocate peaks */
	AllocPeaks();
	
	/* Only show monitor if selected */
	setMonitorVisibility(prefs.show_mon);
	

}

void myCleanup(void)
{
	savePrefs(&prefs);
	freePrefs(&prefs);
	setMonitorVisibility(FALSE);
	XFreeGC(display, blackGC);
	XFreeGC(display, whiteGC);
	XFreeGC(display, blueGC);
	XFreeGC(display, yellowGC);
	XFreeGC(display, dkyellowGC);
	XFreeGC(display, redGC);
	XDestroyWindow(display, window);
	XCloseDisplay(display);

}

void modifyNative16(gpointer *data, gint length)
{
	gint16 *audio = (gint16 *)*data, *ap;
	int peak, pos;
	int i;
	int gr, gf, gn;
	static int pn = -1;
#ifdef STATS
	static int clip = 0;
#endif
	static int clipped = 0;

	if (!peaks)
		return;

	if (pn == -1)
	{
		for (i = 0; i < prefs.buckets; i++)
			peaks[i] = 0;
	}
	pn = (pn + 1)%prefs.buckets;

#ifdef DEBUG
	fprintf(stderr, "modifyNative16(0x%08x, %d)\n",(unsigned int)data,
		length);
#endif

	/* Determine peak's value and position */
	peak = 1;
	pos = 0;

#ifdef DEBUG
	fprintf(stderr, "finding peak(b=%d)\n", pn);
#endif

	ap = audio;
	for (i = 0; i < length/2; i++)
	{
		int val = *ap;
		if (val > peak)
		{
			peak = val;
			pos = i;
		} else if (-val > peak)
		{
			peak = -val;
			pos = i;
		}
		ap++;
	}
	peaks[pn] = peak;

	/* Only draw if needed, of course */
	if (prefs.show_mon)
	{
		/* current amplitude */
		XDrawLine(display, window, whiteGC,
			  pn, 0,
			  pn,
			  127 -
			  (peaks[pn]*gainCurrent >> (GAINSHIFT + 8)));

		/* amplification */
		XDrawLine(display, window, yellowGC,
			  pn,
			  127 - (peaks[pn]*gainCurrent
				 >> (GAINSHIFT + 8)),
			  pn, 127);

		/* peak */
		XDrawLine(display, window, blackGC,
			  pn, 127 - (peaks[pn] >> 8), pn, 127);

		/* clip indicator */
		if (clipped)
			XDrawLine(display, window, redGC,
				  (pn + prefs.buckets - 1)%prefs.buckets,
				  126 - clipped/(length*512),
				  (pn + prefs.buckets - 1)%prefs.buckets,
				  127);
		clipped = 0;

		// target line
		//XDrawPoint(display, window, redGC,
		//         pn, 127 - TARGET/256);
		// amplification edge
		XDrawLine(display, window, dkyellowGC,
			  pn,
			  127 - (peaks[pn]*gainCurrent
				 >> (GAINSHIFT + 8)),
			  pn - 1,
			  127 -
			  (peaks[(pn + prefs.buckets
				  - 1)%prefs.buckets]*gainCurrent
			   >> (GAINSHIFT + 8)));
	}

	for (i = 0; i < prefs.buckets; i++)
	{
		if (peaks[i] > peak)
		{
			peak = peaks[i];
			pos = 0;
		}
	}

	/* Determine target gain */
	gn = (1 << GAINSHIFT)*prefs.target/peak;

	if (gn <(1 << GAINSHIFT))
		gn = 1 << GAINSHIFT;

	gainTarget = (gainTarget *((1 << prefs.gainsmooth) - 1) + gn)
				      >> prefs.gainsmooth;

	/* Give it an extra insignifigant nudge to counteract possible
	** rounding error
	*/

	if (gn < gainTarget)
		gainTarget--;
	else if (gn > gainTarget)
		gainTarget++;

	if (gainTarget > prefs.gainmax << GAINSHIFT)
		gainTarget = GAINSHIFT << GAINSHIFT;


	if (prefs.show_mon)
	{

		/* peak*gain */
		XDrawPoint(display, window, redGC,
			   pn,
			   127 - (peak*gainCurrent
				  >> (GAINSHIFT + 8)));

		// gain indicator
		XFillRectangle(display, window, whiteGC, 0, 128,
			       prefs.buckets, 8);
		XFillRectangle(display, window, redGC,
			       gainTarget*prefs.buckets/(4
							 << GAINSHIFT)
			       - 1,
			       128, 3, 8);
		XDrawLine(display, window, blueGC,
			  prefs.buckets*prefs.target/131072,
			  128, prefs.buckets*prefs.target/131072, 128 + 8);
		XDrawLine(display, window, blackGC, prefs.buckets/4, 128,
			  prefs.buckets/4, 128 + 8);

		XDrawLine(display, window, redGC,
			  gainTarget*prefs.buckets/(4 << GAINSHIFT), 132 - 1,
			  gainTarget*prefs.buckets/(4 << GAINSHIFT), 132 + 1);
		XDrawLine(display, window, whiteGC,
			  gn*prefs.buckets/(4 << GAINSHIFT), 132 - 1,
			  gn*prefs.buckets/(4 << GAINSHIFT), 132 + 1);
		XDrawLine(display, window, blackGC,
			  gn*prefs.buckets/(4 << GAINSHIFT) - 1, 132 - 1,
			  gn*prefs.buckets/(4 << GAINSHIFT) + 1, 132 + 1);
		XDrawLine(display, window, blackGC,
			  gn*prefs.buckets/(4 << GAINSHIFT) + 1, 132 - 1,
			  gn*prefs.buckets/(4 << GAINSHIFT) - 1, 132 + 1);

		// blue peak line
		XDrawLine(display, window, blueGC,
			  0, 127 - (peak >> 8), prefs.buckets,
			  127 - (peak >> 8));
		XFlush(display);
		XDrawLine(display, window, whiteGC,
			  0, 127 - (peak >> 8), prefs.buckets,
			  127 - (peak >> 8));
	}

	/* See if a peak is going to clip */
	gn = (1 << GAINSHIFT)*32768/peak;

	if (gn < gainTarget)
	{
		gainTarget = gn;

		if (prefs.anticlip)
			pos = 0;

	} else
	{
		/* We're ramping up, so draw it out over the whole frame */
		pos = length;
	}

	/* Determine gain rate necessary to make target */
	if (!pos)
		pos = 1;

	gr = ((gainTarget - gainCurrent) << 16)/pos;

	/* Do the shiznit */
	gf = gainCurrent << 16;

#ifdef STATS
	fprintf(stderr, "\rgain = %2.2f%+.2e ",
		gainCurrent*1.0/(1 << GAINSHIFT),
		(gainTarget - gainCurrent)*1.0/(1 << GAINSHIFT));
#endif

	ap = audio;
	for (i = 0; i < length/2; i++)
	{
		int sample;

		/* Interpolate the gain */
		gainCurrent = gf >> 16;
		if (i < pos)
			gf += gr;
		else if (i == pos)
			gf = gainTarget << 16;

		/* Amplify */
		sample = (*ap)*gainCurrent >> GAINSHIFT;
		if (sample < -32768)
		{
#ifdef STATS
			clip++;
#endif
			clipped += -32768 - sample;
			sample = -32768;
		} else if (sample > 32767)
		{
#ifdef STATS
			clip++;
#endif
			clipped += sample - 32767;
			sample = 32767;
		}
		*ap++ = sample;
	}
#ifdef STATS
	fprintf(stderr, "clip %d b%-3d ", clip, pn);
#endif

#ifdef DEBUG
	fprintf(stderr, "\ndone\n");
#endif
}

int myModify(gpointer * data, gint length, AFormat fmt, gint srate, gint nch)
{
	if (fmt == FMT_S16_NE ||
	    (fmt == FMT_S16_LE && G_BYTE_ORDER == G_LITTLE_ENDIAN) ||
	    (fmt == FMT_S16_BE && G_BYTE_ORDER == G_BIG_ENDIAN))
		modifyNative16(data, length);

	return length;
}

void closeAbout(GtkWidget* wg, GtkWidget* win)
{
	
	gtk_widget_destroy(GTK_WIDGET(win));
}

void myAbout(void)
{
	gchar *window_title = "About xmms-compress";
	gchar *about_text =
		"xmms-compress v1.1\n"
                "(c)2002 trikuare studios(http://trikuare.cx)\n\n"
                "Simple dynamic range compressor for transparently\n"
                "keeping the XMMS volume level more or less consistent";

	GtkWidget *about_xmms_compress;
	GtkWidget *vbox1;
	GtkWidget *label1;
	GtkWidget *hseparator1;
	GtkWidget *hbuttonbox1;
	GtkWidget *button1;

	about_xmms_compress = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_container_set_border_width(GTK_CONTAINER(about_xmms_compress),
				       10);
	gtk_window_set_title(GTK_WINDOW(about_xmms_compress), window_title);
	gtk_window_set_wmclass(GTK_WINDOW(about_xmms_compress), "about",
			       "xmms");

	vbox1 = gtk_vbox_new(FALSE, 12);
	gtk_widget_show(vbox1);
	gtk_container_add(GTK_CONTAINER(about_xmms_compress), vbox1);

	label1 = gtk_label_new(about_text);
	gtk_widget_show(label1);
	gtk_box_pack_start(GTK_BOX(vbox1), label1, FALSE, FALSE, 0);

	hseparator1 = gtk_hseparator_new();
	gtk_widget_show(hseparator1);
	gtk_box_pack_start(GTK_BOX(vbox1), hseparator1, TRUE, TRUE, 0);

	hbuttonbox1 = gtk_hbutton_box_new();
	gtk_widget_show(hbuttonbox1);
	gtk_box_pack_start(GTK_BOX(vbox1), hbuttonbox1, TRUE, TRUE, 0);

	button1 = gtk_button_new_with_label("Ok");
	gtk_widget_show(button1);
	gtk_container_add(GTK_CONTAINER(hbuttonbox1), button1);
	GTK_WIDGET_SET_FLAGS(button1, GTK_CAN_DEFAULT);
	gtk_widget_grab_default(button1);

	gtk_signal_connect(GTK_OBJECT(button1), "clicked",
			   GTK_SIGNAL_FUNC(closeAbout),
			   about_xmms_compress);

	gtk_widget_show(GTK_WIDGET(about_xmms_compress));
}

void myPrefs(void)
{
	showPrefs(&prefs);
}

